#include "GraphicsEngine.h"

#define SM_NAME "Default SceneManager"
#define OGRE_UNITSIZE 100

GraphicsEngine::GraphicsEngine()
{
	mStaticObjects = new vector<GameObject*>();
	mBoxes = new vector<GameBox*>();
	mPlayer = NULL;

	createRoot();
	defineResources();
	setupRenderSystem();
	createRenderWindow();
	initializeResourceGroups();
	setupScene();
	setupCEGUI();
	createFrameListener();
}

GraphicsEngine::~GraphicsEngine()
{
	delete mStaticObjects;
	delete mBoxes;

	delete mListener;
	delete mRoot;
}

void GraphicsEngine::createWall(Wall* wall, string texture)
{
	mStaticObjects->push_back(wall);
	
	// Create entity
	Entity* entWall = mRoot->getSceneManager(SM_NAME)->createEntity(wall->getName(), SceneManager::PT_CUBE);
	entWall->setMaterialName(texture);
	entWall->setCastShadows(true);

	// Create sceneNode and add the entity
	SceneNode* nodeWall = mRoot->getSceneManager(SM_NAME)->getRootSceneNode()->createChildSceneNode(wall->getName() + "Node");
	nodeWall->attachObject(entWall);

	// Set position/size of the node according to the wall's data. 
	nodeWall->setPosition(wall->getPos().getX(),wall->getPos().getY(),wall->getPos().getZ());
	nodeWall->scale(wall->getWidth()/OGRE_UNITSIZE,wall->getHeight()/OGRE_UNITSIZE,wall->getDepth()/OGRE_UNITSIZE);
}

void GraphicsEngine::createBox(GameBox* box, string texture)
{
	mBoxes->push_back(box);
	
	// Create entity
	Entity* entBox = mRoot->getSceneManager(SM_NAME)->createEntity(box->getName(), SceneManager::PT_CUBE);
	entBox->setMaterialName(texture);
	entBox->setCastShadows(true);

	// Create sceneNode and add the entity
	SceneNode* nodeBox = mRoot->getSceneManager(SM_NAME)->getRootSceneNode()->createChildSceneNode(box->getName() + "Node");
	nodeBox->attachObject(entBox);

	// Set position/size of the node according to the wall's data. 
	int OgreUnitSize = 100;
	nodeBox->setPosition(box->getPos().getX(),box->getPos().getY(),box->getPos().getZ());
	nodeBox->scale(box->getWidth()/OGRE_UNITSIZE,box->getHeight()/OGRE_UNITSIZE,box->getDepth()/OGRE_UNITSIZE);
}

void GraphicsEngine::createPlayer(Player* player)
{
	SceneManager* mgr = mRoot->getSceneManager(SM_NAME);
	mWin->removeAllViewports();

	// Set up player view of the world
	Camera *cam = mgr->createCamera("Camera");
	cam->setPosition(Vector3(0,0,0));
	cam->pitch(Degree(0));
	cam->yaw(Degree(0));
	cam->setNearClipDistance(1);
	cam->setFarClipDistance(10000);
	Viewport *vp = mWin->addViewport(cam);
	vp->setBackgroundColour(ColourValue(0,0,1));
	// Alter the camera aspect ratio to match the viewport
	cam->setAspectRatio(Real(vp->getActualWidth()) / Real(vp->getActualHeight())); 

	mPlayer = player;

	// Create the camera node and set it's position
	SceneNode* camNode = mgr->getRootSceneNode()->createChildSceneNode("CamNode");
	// Create a sphere around the player
	Entity* entGoal = mRoot->getSceneManager(SM_NAME)->createEntity("PlayerSphere", SceneManager::PT_SPHERE);
	entGoal->setCastShadows(false);
	camNode->attachObject(entGoal);
	camNode->setPosition(player->getPos().getX(),player->getPos().getY()+5,player->getPos().getZ());
	camNode->scale(15.0f/OGRE_UNITSIZE,15.0f/OGRE_UNITSIZE,15.0f/OGRE_UNITSIZE);

	createMiniMap(player->getPos().getX(), player->getPos().getZ());
	mListener->setPlayer(player);
}

void GraphicsEngine::createGoal(Goal* goal)
{
	mStaticObjects->push_back(goal);

	Entity* entGoal = mRoot->getSceneManager(SM_NAME)->createEntity(goal->getName(), SceneManager::PT_SPHERE);
	entGoal->setCastShadows(true);

	SceneNode* nodeGoal = mRoot->getSceneManager(SM_NAME)->getRootSceneNode()->createChildSceneNode(goal->getName() + "Node");
	nodeGoal->attachObject(entGoal);
	nodeGoal->setPosition(goal->getPos().getX(),goal->getPos().getY()+5,goal->getPos().getZ());
	nodeGoal->scale(goal->getRadius()/OGRE_UNITSIZE,goal->getRadius()/OGRE_UNITSIZE,goal->getRadius()/OGRE_UNITSIZE);
}

// creates planes that are probably going to serve as ceiling or ground
void GraphicsEngine::createPlane(std::string name, Vec3 position, int width, int height, int offset, Vec3 orientation, string texture)
{
	Plane plane(Vector3(orientation.getX(),orientation.getY(),orientation.getZ()), offset);
	MeshManager::getSingleton().createPlane(name,
			ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, plane,
			width,height,1,1,true,1,50,50,Vector3::UNIT_Z);
	Entity* ent = mRoot->getSceneManager(SM_NAME)->createEntity(name + "Entity", name);
	
	SceneNode* planeNode = mRoot->getSceneManager(SM_NAME)->getRootSceneNode()->createChildSceneNode(name + "Node");
	planeNode->attachObject(ent);
	planeNode->setPosition(position.getX(), position.getY(), position.getZ());
	ent->setMaterialName(texture);
	ent->setCastShadows(true);
	
}

void GraphicsEngine::createSky(float size, string texture)
{
	Plane plane;
	plane.d = 1000;
	plane.normal = Vector3::NEGATIVE_UNIT_Y;
	mRoot->getSceneManager(SM_NAME)->setSkyPlane(true, plane, texture, size, size*0.02667, true, 1.5f, 150, 150);
}

void GraphicsEngine::createRoot()
{
	mRoot = new Root();
}
   
void GraphicsEngine::defineResources()
{
	String secName, typeName, archName;
	ConfigFile cf;
	cf.load("resources.cfg");

	ConfigFile::SectionIterator seci = cf.getSectionIterator();
	while (seci.hasMoreElements())
	{
		secName = seci.peekNextKey();
		ConfigFile::SettingsMultiMap *settings = seci.getNext();
		ConfigFile::SettingsMultiMap::iterator i;
		for (i = settings->begin(); i != settings->end(); ++i)
		{
			typeName = i->first;
			archName = i->second;
			ResourceGroupManager::getSingleton().addResourceLocation(archName, typeName, secName);
		}
	}
}
    
void GraphicsEngine::setupRenderSystem()
{
	if (!mRoot->restoreConfig() && !mRoot->showConfigDialog())
	throw Exception(52, "User canceled the config dialog!", "Application::setupRenderSystem()");
}
    
void GraphicsEngine::createRenderWindow()
{
	mRoot->initialise(false);
	HWND hWnd = 0;
	NameValuePairList misc;
	misc["externalWindowHandle"] = StringConverter::toString((int)hWnd);
	mWin = mRoot->createRenderWindow("Main RenderWindow", 800, 600, true, &misc);
}

void GraphicsEngine::initializeResourceGroups()
{
	TextureManager::getSingleton().setDefaultNumMipmaps(5);
	ResourceGroupManager::getSingleton().initialiseAllResourceGroups();
}

void GraphicsEngine::setupScene()
{
	SceneManager *mgr = mRoot->createSceneManager(ST_GENERIC, SM_NAME);
	mgr->setAmbientLight( ColourValue( 1, 1, 1 ) );
	mgr->setShadowTechnique(SHADOWTYPE_STENCIL_ADDITIVE);
}

void GraphicsEngine::createMiniMap(int x, int z)
{
	SceneManager* mgr = mRoot->getSceneManager(SM_NAME);

	//Set up map view of the world	
	Camera* cam = mgr->createCamera("MapCamera");
	cam->setPosition(x, 900, z);
	cam->pitch(Degree(-90));
	Viewport* vp = mWin->addViewport(cam, 1,0.75,0.75,0.2,0.2);
	vp->setOverlaysEnabled(false);
	cam->setAspectRatio(Real(vp->getActualWidth()) / Real(vp->getActualHeight()));
}

void GraphicsEngine::setupCEGUI()
{
	SceneManager *mgr = mRoot->getSceneManager(SM_NAME);

	// CEGUI setup
	mRenderer = new CEGUI::OgreCEGUIRenderer(mWin, Ogre::RENDER_QUEUE_OVERLAY, false, 3000, mgr);
	mSystem = new CEGUI::System(mRenderer);

	// Other CEGUI setup here.
	CEGUI::SchemeManager::getSingleton().loadScheme((CEGUI::utf8*)"TaharezLookSkin.scheme");
}

void GraphicsEngine::createFrameListener()
{
	size_t windowHnd = 0;
	std::ostringstream windowHndStr;
	OIS::ParamList pl;

	mWin->getCustomAttribute("WINDOW", &windowHnd);
	windowHndStr << windowHnd;
	pl.insert(std::make_pair(std::string("WINDOW"), windowHndStr.str()));

	mListener = new InputListener(pl, mRoot->getSceneManager(SM_NAME));
	mListener->setupRoom(mStaticObjects, mBoxes);
	mRoot->addFrameListener(mListener);
}

void GraphicsEngine::reset()
{
	mListener->setupRoom(mStaticObjects, mBoxes);
}

/**
 *	Starts the renderloop. It return wether the game should continue to the next room (true) or
 *	end the game (false)
 */
bool GraphicsEngine::startRenderLoop()
{
	while (mRoot->renderOneFrame())
	{
		// Make sure you call the display update function:
		WindowEventUtilities::messagePump();	
		
		// Render one frame. After a frame has ended it could be possible that the player has finished the room
		if (mListener->getRoomEnd())
		{
			clearScene();
			return true;
		}
	}
	return false;
}

void GraphicsEngine::clearScene()
{
	// destroy visible objects
	for (unsigned int i = 0; i < mStaticObjects->size(); i++)
	{
		mRoot->getSceneManager(SM_NAME)->destroyEntity(mStaticObjects->at(i)->getName());
		mRoot->getSceneManager(SM_NAME)->destroySceneNode(mStaticObjects->at(i)->getName() + "Node");
	}
	
	for (unsigned int i = 0; i < mBoxes->size(); i++)
	{
		mRoot->getSceneManager(SM_NAME)->destroyEntity(mBoxes->at(i)->getName());
		mRoot->getSceneManager(SM_NAME)->destroySceneNode(mBoxes->at(i)->getName() + "Node");
	}
	
	mRoot->getSceneManager(SM_NAME)->destroyEntity("groundEntity");	
	mRoot->getSceneManager(SM_NAME)->destroySceneNode("groundNode");
	MeshManager::getSingleton().remove(MeshManager::getSingleton().getByName("ground"));

	mRoot->getSceneManager(SM_NAME)->destroyEntity("PlayerSphere");	

	// destroy invisible objects
	mRoot->getSceneManager(SM_NAME)->destroyCamera("Camera");
	mRoot->getSceneManager(SM_NAME)->destroySceneNode("CamNode");

	mRoot->getSceneManager(SM_NAME)->destroyCamera("MapCamera");

	mStaticObjects->clear();
	mBoxes->clear();
}