#include "InputListener.h"
#include "Functions.h"

#include <CEGUI/CEGUISystem.h>
#include <CEGUI/CEGUISchemeManager.h>
#include <CEGUI/CEGUIWindow.h>
#include <CEGUI/CEGUIWindowManager.h>

#include <OgreTextAreaOverlayElement.h>
#include <OgreCEGUIRenderer.h>

#include <cmath>

InputListener::InputListener(OIS::ParamList pl, SceneManager *sceneMgr)
{
	mSceneMgr = sceneMgr;
	mInputManager = OIS::InputManager::createInputSystem(pl);
	try
	{
		mKeyboard = static_cast<OIS::Keyboard*>(mInputManager->createInputObject(OIS::OISKeyboard, true));
		mMouse = static_cast<OIS::Mouse*>(mInputManager->createInputObject(OIS::OISMouse, true));
	}
	catch (const OIS::Exception &e)
	{
		throw new Exception(42, e.eText, "Application::setupInputSystem");
	}
	
	mMouse->setEventCallback(this);
	mKeyboard->setEventCallback(this);		

	// set the rotation and move speed
	mRotate = 0.13;
	mMove = 400;

	// continue rendering
	mContinue = true;

	mDirection = Vector3::ZERO;

	overlayTime = 0;
	topOverlay = NULL;
	bigOverlay = NULL;
}

InputListener::~InputListener()
{
	if (mInputManager)
	{
		mInputManager->destroyInputObject( mMouse );
		mInputManager->destroyInputObject( mKeyboard );
		
		OIS::InputManager::destroyInputSystem(mInputManager);	
	}
}

void InputListener::setPlayer(Player* player)
{
	mPlayer = player;

	// Create the camera's top node (which will only handle position).
	mCamNode = mSceneMgr->getSceneNode("CamNode");

	// Create the camera's yaw node as a child of camera's top node.
	mCamYawNode = mCamNode->createChildSceneNode();

	// Create the camera's pitch node as a child of camera's yaw node.
	mCamPitchNode = mCamYawNode->createChildSceneNode();
	mCamPitchNode->attachObject(mSceneMgr->getCamera("Camera"));
}

bool InputListener::frameStarted(const FrameEvent& evt)
{
	if(mMouse)
		mMouse->capture();
	if(mKeyboard) 
		mKeyboard->capture();
	
	moveCamera(evt.timeSinceLastFrame);
	updateObjects(evt.timeSinceLastFrame);

	showTopOverlay("Number of pushes: " + Functions::toString(noOfPushes));
	return mContinue;
}

void InputListener::updateObjects(Real timeSinceLastFrame)
{
	for (unsigned int i=0; i < mBoxes->size(); i++)
	{
		Vec3 updatedPos = mBoxes->at(i)->moveToNewPos(timeSinceLastFrame);
		SceneNode* boxNode = mSceneMgr->getSceneNode(mBoxes->at(i)->getName() + "Node");
		boxNode->setPosition(updatedPos.getX(), updatedPos.getY(), updatedPos.getZ());
	}
}

void InputListener::restartRoom()
{
	noOfPushes = 0;
	// Set the player to the old position
	if (mPlayer != NULL)
	{
		Vec3 updatedPos = mPlayer->toStartPos();
		mCamNode->setPosition(updatedPos.getX(), updatedPos.getY(), updatedPos.getZ());
	}
	for (unsigned int i=0; i < mBoxes->size(); i++)
	{
		Vec3 updatedPos = mBoxes->at(i)->toStartPos();
		SceneNode* boxNode = mSceneMgr->getSceneNode(mBoxes->at(i)->getName() + "Node");
		boxNode->setPosition(updatedPos.getX(), updatedPos.getY(), updatedPos.getZ());
	}
}

/**
 *	Moves the camera depending on the yaw of the camere and the direction it's heading to
 */
void InputListener::moveCamera(Real timeSinceLastFrame)
{
	// move player camera
	Vector3 oldPos = mCamNode->getPosition();
	mCamNode->translate(mCamYawNode->getOrientation() * mDirection * timeSinceLastFrame, Node::TS_LOCAL);
	Vector3 pos = mCamNode->getPosition();

	if (!moveToPos(pos,oldPos) || moveBox(pos,oldPos))
		mCamNode->setPosition(oldPos.x, oldPos.y, oldPos.z);

	// move map camera
	Camera* cam = mSceneMgr->getCamera("MapCamera");
	cam->setPosition(mCamNode->getPosition().x, cam->getPosition().y, mCamNode->getPosition().z);
}

/**
 *	Checks if the player pushes a box and if the box can be pushed. Return wether a box has been hit
 */
bool InputListener::moveBox(Vector3 pos, Vector3 oldPos)
{
	// check if new pos is in a gameBox and if so move the box
	GameBox* box = isBoxFree(pos);
	if (box != NULL)
	{
		if (!box->isMoving())
		{
			float dx = oldPos.x - box->getX();
			float dz = oldPos.z - box->getZ();

			Vector3 newPos;

			// Make sure the Box will be moving
			if (box->containsPos(oldPos.x,oldPos.y, pos.z))
			{
				// sign says wether the box should go left or right
				int sign = -(dz/abs(dz));
				newPos = Vector3(box->getX(), box->getY(), box->getZ() +  sign * box->getDepth());
			}
			else
			{
				// sign says wether the box should go up or down
				int sign = -(dx/abs(dx));
				newPos = Vector3(box->getX() +  sign * box->getWidth(), box->getY(), box->getZ());
			}
			if (isPosFree(newPos) == NULL && isBoxFree(newPos) == NULL)
			{
				box->setNewPos(newPos.x, newPos.y, newPos.z);			
				noOfPushes++;
			}
		}
		return true;
	}
	return false;
}
/**
 *	Tries to move the camera node to a given position. Returns wether it succeeded
 *
 *	If it returns false, the camera should be set back to it's old position
 */
bool InputListener::moveToPos(Vector3 pos, Vector3 oldPos)
{
	// check if new pos in in a wall or in the goal
	GameObject* object = isPosFree(pos);
	Wall* hitWall = dynamic_cast<Wall*>(object);
	if (hitWall != NULL)
	{
		Vector3 newPos;
		// new pos is in a wall -> adjust position
		if (abs(pos.x - hitWall->getX()) < abs(pos.z - hitWall->getZ()))
		{
			newPos = Vector3(pos.x, oldPos.y, oldPos.z);
		}
		else
		{
			newPos = Vector3(oldPos.x, oldPos.y, pos.z);
		}
		if (isBoxFree(newPos) == NULL)
		{
			GameObject* obj = isPosFree(newPos);
			Goal* goal = dynamic_cast<Goal*>(obj);
			if (goal != NULL) // you are on the finish
			{
				endRoom();				
				return false;
			}
			else
			{
				if (obj == NULL) // You are in a free position
				{
					mCamNode->setPosition(newPos.x, newPos.y, newPos.z);
					return true;
				}
				else
					return false;
			}
		}
		else
			return false;
	}
	else
	{
		Goal* goal = dynamic_cast<Goal*>(object);
		if (goal != NULL)
		{
			endRoom();
			return false;
		}
		else
			return true;
	}
}

bool InputListener::getRoomEnd()
{
	return bRoomEnd;
}

void InputListener::endRoom()
{
	showBigOverlay("You finished the room with " + Functions::toString(noOfPushes) + " pushes!\n\nPress esc to quit or space to go\nto the next level.");
}

void InputListener::setupRoom(vector<GameObject*>* staticObjects, vector<GameBox*>* boxes)
{
	noOfPushes = 0;
	mStaticObjects = staticObjects;
	mBoxes = boxes;
	mPlayer = NULL;
	bRoomEnd = false;
	if(bigOverlay != NULL)
		bigOverlay->hide();
	if(topOverlay != NULL)
		topOverlay->hide();
}

GameObject* InputListener::isPosFree(Vector3 pos)
{
	for (unsigned int i=0; i < mStaticObjects->size(); i++)
	{
		// Check if new pos is in a wall
		if (mStaticObjects->at(i)->containsPos(pos.x,pos.y,pos.z))
		{
			return mStaticObjects->at(i);
		}
	}
	return NULL;
}
GameBox* InputListener::isBoxFree(Vector3 pos)
{
	for (unsigned int i=0; i < mBoxes->size(); i++)
	{
		// Check if new pos is in a wall
		if (mBoxes->at(i)->containsPos(pos.x,pos.y,pos.z))
		{
			return mBoxes->at(i);
		}
	}
	return NULL;
}

bool InputListener::mouseMoved(const OIS::MouseEvent &e) 
{
	// change the direction of the camera according to the mouse movement
	mCamYawNode->yaw(Degree(-mRotate * e.state.X.rel), Node::TS_LOCAL);
	mCamPitchNode->pitch(Degree(-mRotate * e.state.Y.rel), Node::TS_LOCAL);

	return true;
}

bool InputListener::mousePressed(const OIS::MouseEvent &e, OIS::MouseButtonID id)
{
	return true;
}

bool InputListener::mouseReleased(const OIS::MouseEvent &e, OIS::MouseButtonID id)
{
	return true;
}

// shows a small text in left upper corner of the screen
void InputListener::showTopOverlay(String text)
{
	if(!topOverlay)
	{
		OverlayManager& overlayManager = OverlayManager::getSingleton();

		// Create a panel
		OverlayContainer* panel = static_cast<OverlayContainer*>(
			overlayManager.createOverlayElement("Panel", "PanelName"));
		panel->setMetricsMode(Ogre::GMM_PIXELS);
		panel->setPosition(10, 10);
		panel->setDimensions(100, 100);

		// Create a text area
		TextAreaOverlayElement* textArea = static_cast<TextAreaOverlayElement*>(
			overlayManager.createOverlayElement("TextArea", "TextAreaName"));
		textArea->setMetricsMode(Ogre::GMM_PIXELS);
		textArea->setPosition(0, 0);
		textArea->setDimensions(100, 100);
		textArea->setCaption(text);
		textArea->setCharHeight(16);
		textArea->setFontName("StarWars");
		textArea->setColourBottom(ColourValue(0.3, 0.5, 0.3));
		textArea->setColourTop(ColourValue(0.5, 0.7, 0.5));

		// Create an overlay, and add the panel
		topOverlay = overlayManager.create("OverlayName");
		topOverlay->add2D(panel);

		// Add the text area to the panel
		panel->addChild(textArea);
	}
	else
	{
		topOverlay->getChild("PanelName")->getChild("TextAreaName")->setCaption(text);
	}

	// Show the overlay
	topOverlay->show();
}

// shows a big text
void InputListener::showBigOverlay(String text)
{
	if(!bigOverlay)
	{
		OverlayManager& overlayManager = OverlayManager::getSingleton();

		// Create a panel
		OverlayContainer* panel = static_cast<OverlayContainer*>(
			overlayManager.createOverlayElement("Panel", "BigPanelName"));
		panel->setMetricsMode(Ogre::GMM_PIXELS);
		panel->setPosition(100, 100);
		panel->setDimensions(600, 100);
		panel->setMaterialName("Core/StatsBlockBorder/Up");

		// Create a text area
		TextAreaOverlayElement* textArea = static_cast<TextAreaOverlayElement*>(
			overlayManager.createOverlayElement("TextArea", "BigTextAreaName"));
		textArea->setMetricsMode(Ogre::GMM_PIXELS);
		textArea->setPosition(65, 17);
		textArea->setDimensions(100, 100);
		textArea->setCaption(text);
		textArea->setCharHeight(16);
		textArea->setFontName("StarWars");
		textArea->setColourBottom(ColourValue(0.5, 0.2, 0.3));
		textArea->setColourTop(ColourValue(0.7, 0.3, 0.4));

		// Create an overlay, and add the panel
		bigOverlay = overlayManager.create("BigOverlayName");
		bigOverlay->add2D(panel);

		// Add the text area to the panel
		panel->addChild(textArea);
	}
	else
	{
		bigOverlay->getChild("BigPanelName")->getChild("BigTextAreaName")->setCaption(text);
	}

	// Show the overlay
	bigOverlay->show();
}

void InputListener::zoomMiniMap(int amount)
{
	Camera* cam = mSceneMgr->getCamera("MapCamera");
	cam->setPosition(cam->getPosition().x, cam->getPosition().y+amount, cam->getPosition().z);
}


// KeyListener
bool InputListener::keyPressed(const OIS::KeyEvent &e)
{
	switch (e.key)
	{
		case OIS::KC_UP:
		case OIS::KC_W:
			mDirection.z = -mMove;
			break;

		case OIS::KC_DOWN:
		case OIS::KC_S:
			mDirection.z = mMove;
			break;

		case OIS::KC_LEFT:
		case OIS::KC_A:
			mDirection.x = -mMove;
			break;

		case OIS::KC_RIGHT:
		case OIS::KC_D:
			mDirection.x = mMove;
		break;

		case OIS::KC_ESCAPE: 
			mContinue = false;
			break;

		case OIS::KC_SPACE:
			bRoomEnd = true;
			break;

		case OIS::KC_ADD:
			zoomMiniMap(-50);
			break;
		case OIS::KC_SUBTRACT:
			zoomMiniMap(50);
			break;

		default:
			break;
	}
	return mContinue;

}

bool InputListener::keyReleased(const OIS::KeyEvent &e)
{
	switch (e.key)
	{
		case OIS::KC_UP:
		case OIS::KC_W:
			mDirection.z = 0;
			break;

		case OIS::KC_DOWN:
		case OIS::KC_S:
			mDirection.z = 0;
			break;

		case OIS::KC_LEFT:
		case OIS::KC_A:
			mDirection.x = 0;
			break;

		case OIS::KC_RIGHT:
		case OIS::KC_D:
			mDirection.x = 0;
			break;
		
		case OIS::KC_R:
			restartRoom();
			break;

		default:
			break;
	} // switch
	return true;

}